/****************************************************************
 *                                                              *
 *  (C)  The TriTechnologies Company, 1991                      *
 *  	      	   	     	       	    	      	   	*
 *       This software is furnished under a license and may     *
 *       be used and copied only in accordance with the terms   *
 *       of such license and with the inclusion of the above    *
 *       copyright notice.  this software or any other copies   *
 *       thereof may not be provided or otherwise made avail-   *
 *       able to any other person.  No title to and ownership   *
 *       of the software is hereby transferred.                 *
 *                                                              *
 *       The information in this software is subject to change  *
 *       without notice and should not be construed as a        *
 *       commitment by The TriTechnologies Company.             *
 ****************************************************************
 *
 * Facility:
 *
 *     rwlocks.c
 *
 * Abstract:
 *
 *       rwlocks.c contains the example routines illustrating shared
 *       and exclusive locking.
 *
 *       These routines illustrate the use of the broadcast facility
 *       provided by PCthreads and provide a general introduction to the
 *       use of condition variables.
 */

#include "pcthread.h"

static int active_threads, 
           waiting_readers = 0, 
           waiting_writers = 0;

static pcthread_mutex_t      mutex;
static pcthread_cond_t       cv_write, cv_read;

void
lock_shared()
{
    (void) pcthread_mutex_lock( (pcthread_mutex_t *) &mutex );

         waiting_readers += 1;

         /*
          * Check to see if there are any writers waiting to gain access
          * to the critical section.  If there are, then we will arrange
          * for them to proceed.
          */

         if (waiting_writers > 0)
         {
              /*
               * There are 1 or more reader threads waiting to gain
               * access to this critical region.  We signal the
               * first writer.  Note that since we hold the lock
               * we eliminate the possibility of a higher priority writer 
               * thread preempting us before we issue the condition wait.
               */
               
              pcthread_cond_signal( (pcthread_cond_t *) &cv_write);
              pcthread_cond_wait ( (pcthread_cond_t *) &cv_read, 
                                   (pcthread_mutex_t *) &mutex);
         }

         /*
          * If active_threads is less than zero, there is a writer in the 
          * critical section.  We must block until the writer is complete.
          */

         while ( active_threads < 0 )
              pcthread_cond_wait ( (pcthread_cond_t *) &cv_read, 
                                   (pcthread_mutex_t *) &mutex);

         waiting_readers -= 1;
         active_threads += 1;

    (void) pcthread_mutex_unlock( (pcthread_mutex_t *) &mutex );

    if ( waiting_readers > 0)
         pcthread_cond_signal( (pcthread_cond_t *) &cv_read );

}

void
unlock_shared()
{
    boolean_t dosignal = false;

    (void) pcthread_mutex_lock( (pcthread_mutex_t *) &mutex );

    /*
     * We begin by decrementing the active thread counter.  If the
     * the number of active threads is equal to zero, then we're the
     * the last reader through the critical section.  We make a note
     * of that fact by setting the dosignal flag to true.
     */
          
     active_threads -= 1;

     if ( active_threads == 0 && waiting_writers > 0 )
         dosignal = true;

    (void) pcthread_mutex_unlock( (pcthread_mutex_t *) &mutex );

    if ( dosignal == true )
         (void) pcthread_cond_signal( (pcthread_cond_t *) &cv_write);
}

void
lock_exclusive()
{
    (void) pcthread_mutex_lock( (pcthread_mutex_t *) &mutex );

         /*
          * First, we increment the waiting_writers count, then we check
          * to see if their are any active threads within the critical
          * section.  If there are, we wait on the cv_write condition
          * variable.
          */
          
         waiting_writers += 1;

         while ( active_threads != 0 )
              pcthread_cond_wait( (pcthread_cond_t *) &cv_write, 
                                  (pcthread_mutex_t *) &mutex );

         /*
          * We're here because the number of active threads is zero.
          * Decrement the waiting writers counter, set the active threads
          * flag and unlock the mutex.
          */
          
         waiting_writers -= 1;
         active_threads = -1;

    (void) pcthread_mutex_unlock( (pcthread_mutex_t *) &mutex );
}

void
unlock_exclusive()
{

    (void) pcthread_mutex_lock( (pcthread_mutex_t *) &mutex );

         /*
          * Set the active thread counter back to zero -- Remember, we're
          * a writer thread and there must only be one of us.  Therefore,
          * when we unlock the exclusive lock, it makes sense to set the
          * active thread counter to zero.
          */
          
         active_threads = 0;

         /*
          * If there are other writers, signal the cv_write condition variable
          * because we only want 1 writer at a time.  Howver, if there are
          * no writers and a 1 or more readers, then broadcast to the
          * cv_read condition variable.
          */

         if ( waiting_writers > 0)
              (void) pcthread_cond_signal( (pcthread_cond_t *) &cv_write);
         else
              if ( waiting_readers > 0 )
                   pcthread_cond_broadcast( (pcthread_cond_t *) &cv_read );


    (void) pcthread_mutex_unlock( (pcthread_mutex_t *) &mutex );
}

int
create_rwlocks()
{
    int  status = 0;

    status = pcthread_mutex_init( (pcthread_mutex_t *) &mutex,
                                   pcthread_mutexattr_default );
    if( status < 0 )
         return( status );

    status = pcthread_cond_init( (pcthread_cond_t *) &cv_write,
                                  pcthread_condattr_default );
    if( status < 0 )
         return( status );
         
    status = pcthread_cond_init( (pcthread_cond_t *) &cv_read,
                                  pcthread_condattr_default );

    return( status );
}